package io.tswf.groovy.bettergroovy.transforms.annotationgstrings

import org.codehaus.groovy.ast.ASTNode
import org.codehaus.groovy.ast.AnnotationNode
import org.codehaus.groovy.ast.ClassNode
import org.codehaus.groovy.ast.expr.ConstantExpression
import org.codehaus.groovy.ast.expr.GStringExpression
import org.codehaus.groovy.control.CompilePhase
import org.codehaus.groovy.control.SourceUnit
import org.codehaus.groovy.transform.AbstractASTTransformation
import org.codehaus.groovy.transform.GroovyASTTransformation
import org.codehaus.groovy.transform.GroovyASTTransformationClass
import org.codehaus.groovy.vmplugin.VMPluginFactory

import java.lang.annotation.Retention
import java.lang.annotation.RetentionPolicy

@GroovyASTTransformationClass(classes = GStringInAnnotationsPreprocessor)
@Retention(RetentionPolicy.SOURCE)
@interface PreprocessGStringsForAnnotations {

}

@GroovyASTTransformation(phase = CompilePhase.CONVERSION)
class GStringInAnnotationsPreprocessor extends AbstractASTTransformation{
    public static final String DELAYED_GSTIRNG_META_KEY = "${GStringInAnnotationsPreprocessor.name}-cached-g-string"

    @Override
    void visit(ASTNode[] nodes, SourceUnit source) {
        init(nodes, source)
        def cn = nodes.find {it instanceof ClassNode} as ClassNode
        cn.annotations.forEach {delayGStringsInAnnotation(it)}

        cn.allDeclaredMethods.forEach {mn ->
            mn.annotations.forEach {delayGStringsInAnnotation(it)}

            mn.parameters.each {pn ->
                pn.annotations.forEach {delayGStringsInAnnotation(it)}
            }
        }

        cn.declaredConstructors.forEach {mn ->
            mn.annotations.forEach {delayGStringsInAnnotation(it)}

            mn.parameters.each {pn ->
                pn.annotations.forEach {delayGStringsInAnnotation(it)}
            }
        }

        cn.fields.forEach {mn ->
            mn.annotations.forEach {delayGStringsInAnnotation(it)}
        }
    }

    void delayGStringsInAnnotation(AnnotationNode annotation) {
        def members = new HashMap<>(annotation.members)
        for (def memberEntry : members) {
            def memberValueExpr = memberEntry.value
            if (memberValueExpr instanceof GStringExpression) {
                def memberGStringValue = memberValueExpr as GStringExpression

                def constant = new ConstantExpression("GString Will Be Here!")
                constant.putNodeMetaData(DELAYED_GSTIRNG_META_KEY, memberGStringValue)
                annotation.members[memberEntry.key] = constant
            }
        }

        VMPluginFactory.getPlugin().configureAnnotation(annotation)
    }
}
